"""
This Module handles the generation of results in the CLI, based on the ticker passed as the argument to the object of the Main class. 
The execution begins by retrieving the financial data from yfinace library which is handled in the data module of stockDL.
After the data is collected the first trading day of each month is stored in a list by the date_calculation() function in the data module.
The data is then preprocessed by the preprocessor module which creates a window to run the predictions on.
It also comprises min-max scaling which reduces sudden highs and lows in the data which would have resulted in anomalies. 
The training module trains the model created in the model module on the data retrieved by the preprocessing module. 
The plots module helps in plotting the necessary graphs for better visualisation for EDA.
The result module uses the calculation and market modules to run the necessary calculations on the predictions,
and generate net and the gross yield on the predictions obtained. 
"""
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
from . import preprocessing, train, plots, results
trained = False
class Main:
    def __init__(self, ticker, saved):
        '''Ticker is a unique symbol for every stock in the share market'''
        self.ticker = ticker
        '''this module helps in loading the data by calling an instance of data.py module and also helps in data preprocessing.'''
        self.saved = saved
        self.data_preprocessor = preprocessing.data_preprocessing(ticker)
        '''This variable stores the monthly data for analysis by the model'''
        self.df_monthly = self.data_preprocessor.df_monthly
        '''This instance of train module will help in preventing the training to occur more than once thus reducing the probability of overfitting.''' 
        self.train = train.Training(ticker, saved)
        '''Trains the data on the defined models'''
        #self.train.train_model()
        '''Creates an instance of the plots module for better visualisation of training and validation data'''
        self.plots = plots.Plots(ticker)
        '''Generates a comparison plot of the 4 methos used'''
        #self.plots.comparison_plots()
        '''An instance of the result module to calculate and process the final predictions on the data by the trained model.'''
        self.results = results.Results(ticker)
        '''Stores the final result in a pandas dataframe'''
        self.result = self.results.result
        '''Converts the pandas dataframe to JSON for better utilisation by web developers'''
        self.result_json = self.result.to_json(orient="split")
        print(self.result)



'''
This module helps in data collection from Yahoo Finance API with the help of yfinace library. 
The stock data is loaded from its unique stock symbol, also called ticker. 
After the data is loaded, we drop the axis which isn't being used by us in the library.
After dropping the unnecessary axis we calculate and store the first days of each trading month in a list. 
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
import datetime
import yfinance as yf
import tensorflow as tf


''' 
If the GPU is available, 
It will be restricted from consuming all memory 
else, these lines are ignored. 
'''
gpus = tf.config.experimental.list_physical_devices('GPU')
if len(gpus) > 0:
    tf.config.experimental.set_memory_growth(gpus[0], True)
    print("Using GPU")


class Data_Loader():
    def __init__(self, ticker):
        self.ticker = ticker
        self.df = self.load_data()
        self.start_date = self.df.index[0].to_pydatetime()
        self.end_date = datetime.date.today()
        ''' Stores the first day of each month from the starting date of stock to the day of executing the library '''
        self.first_days = self.date_calculations()
    
    ''' This function loads the data of a particular stock on the basis of the ticker provided by the user.  '''
    def load_data(self):
        stock = yf.Ticker(self.ticker)
        stock.info
        # get historical market data
        df = stock.history(period="max")
        df.drop("Dividends", axis=1, inplace=True)
        df.drop("Stock Splits", axis=1, inplace=True)
        return df
    
    ''' This function calculates the first trading day for each month from the first day i.e. the opening day of stock, 
    and the last day i.e. the day of executing the library. '''
    def date_calculations(self):
        start_year = self.start_date.year
        start_month = self.start_date.month
        end_year = self.end_date.year
        end_month = self.end_date.month

        first_days = []  

        for month in range(start_month, 13):
            first_days.append(min(self.df.loc[str(start_year)+"-"+str(month)].index))

        for year in range(start_year+1, end_year):
            for month in range(1, 13):
                first_days.append(min(self.df.loc[str(year)+"-"+str(month)].index))
        
        for month in range(1, end_month+1):
            first_days.append(min(self.df.loc[str(end_year)+"-"+str(month)].index))

        return first_days


'''
This module stores the stock market variables related to the stock ticker, 
This module must be run after training the model by calling the train_model() function in the train module.
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
import numpy as np
from . import train
class Market():
    def __init__(self, ticker):
        self.train = train.Training(ticker, "no")
        self.w_lstm = np.diff(self.train.y_pred_lstm.reshape(self.train.y_pred_lstm.shape[0]), 1)
        ''' This stores the predictions of the lstm model and reshapes it. '''
        self.v_lstm = np.maximum(np.sign(self.w_lstm), 0)
        ''' Stores the sign of the lstm predictions as +ve, -ve or zero which implies profit, loss and no profit no loss in the trade. '''
        self.w_mix = np.diff(self.train.y_pred_mix.reshape(self.train.y_pred_mix.shape[0]), 1)
        ''' This stores the predictions of the mix model (conv1D + LSTM) and reshapes it. '''
        self.v_mix = np.maximum(np.sign(self.w_mix), 0)
        ''' Stores the sign of the mix model predictions as +ve, -ve or zero which implies profit, loss and no profit no loss in the trade. '''


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="stockDL",
    version="0.2.0",
    description="Predicts the Gross Yield, Annual Yield and Net Yield of a user given stock ticker.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ashishpapanai/stockDL",
    author="Ashish Papanai",
    author_email="ashishpapanai00@gmail.com",
    py_modules=["calculations", "data", "main", "market", "models", "plots", "preprocessing", "results", "train"],
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["stockDL"],
    include_package_data=True,
    install_requires=["pandas", "numpy", "matplotlib", "keras", "tensorflow", "yahoo-finance", "yfinance"],
    entry_points={
        "console_scripts": [
            "stocksDL=stocksDL.__main__:main",
        ]
    },
)



'''
This module stores the brain of the library which is the Deep Learning model.
The two Deep Learning strategies are defined in their respective methods. 
This module requires to run the preprocessing module so that the model gets the data to work on. 
'''
import os  
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  
from keras.models import Sequential
from keras.optimizers import Adam
from keras.layers import Dense, Dropout, Conv1D, LSTM

from . import preprocessing


class Models():
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.lstm_model = self.LSTM_Model(self.preprocessing.window+1, 8)
        self.mix_lstm_model = self.Mix_LSTM_Model(self.preprocessing.window+1, 8)
    '''
    The LSTM model is based on the LSTM Network which is a modified and better implementation of RNN, 
    LSTM are free from the vanishing gradients problem and so we use them for time-series predictions. 

    LSTM Model: 
    The first deep learning model comprises the following layers:
    1. LSTM layer with 300 nodes
    2. Dropout layer: which discards 50% nodes to reduce overfitting
    3. LSTM layer 2 with 200 nodes to improve model
    4. Dropout layer 2: which discards 50% nodes to reduce overfitting
    5. Dense layer with 100 nodes to reduce the number of nodes 
    6. Final Dense layer with 1 node [This layer is used as an output layer] for regression. 
    '''
    def LSTM_Model(self, window, features):
        model = Sequential()
        model.add(LSTM(300, input_shape=(window, features), return_sequences=True))
        model.add(Dropout(0.5))
        model.add(LSTM(200,  return_sequences=False))
        model.add(Dropout(0.5))
        model.add(Dense(100, kernel_initializer='uniform', activation='relu'))
        model.add(Dense(1, kernel_initializer='uniform', activation='relu'))
        model.compile(loss='mse', optimizer=Adam(learning_rate=0.001))

        return model

    '''
    Mix LSTM Model is an improved approach with 1D convolution layer as the LSTM layers are computationally costly, 
    We use the Conv1D layer to learn the features in the data at much less computational cost. 

    
    MIX LSTM MODEL:
    Same as the original LSTM model but with improved and computationally cheap feature extraction in data by two 1D Convolution

    '''
    def Mix_LSTM_Model(self, window, features):

        model = Sequential()
        model.add(Conv1D(input_shape=(window, features), filters=32,
                        kernel_size=2, strides=1, activation='relu', padding='same'))
        model.add(Conv1D(filters=64, kernel_size=2, strides=1,
                        activation='relu', padding='same'))
        model.add(LSTM(300, return_sequences=True))
        model.add(Dropout(0.5))
        model.add(LSTM(200,  return_sequences=False))
        model.add(Dropout(0.5))
        model.add(Dense(100, kernel_initializer='uniform', activation='relu'))
        model.add(Dense(1, kernel_initializer='uniform', activation='relu'))
        model.compile(loss='mse', optimizer=Adam(learning_rate=0.001))

        return model
